import re
from datetime import datetime
from json import dumps as json_dumps

from django import template

register = template.Library()


@register.filter(is_safe=True)
def digitcomma(value):
    """
    Convert an integer to a string containing commas every three digits.
    For example, 3000 becomes '3,000' and 45000 becomes '45,000'.

    - 基于 django.contrib.humanize 重新实现了一个 tempalte filter
    - 解决 USE_L10N = True 后这个 intcomma filter 不工作的问题
    - 使用 digitcomma 这个新名字已避免与 humanize 内的发生冲突
    """
    orig = str(value)
    new = re.sub(r"^(-?\d+)(\d{3})", r"\g<1>,\g<2>", orig)
    if orig == new:
        return new
    else:
        return digitcomma(new)


@register.filter
def json(value):
    return json_dumps(value)


@register.filter("startswith")
def startswith(text, starts):
    if isinstance(text, str):
        return text.startswith(starts)
    return False


@register.filter("datetime_ss")
def datetime_simple_short(value: datetime):
    if value is None:
        return ""

    value = value.astimezone()
    return value.strftime("%m/%d %H:%M")



#!/usr/bin/env python


import os.path


# https://github.com/getsentry/raven-python/blob/master/raven/versioning.py
def fetch_git_sha(path, head=None):
    """
    >>> fetch_git_sha(os.path.dirname(__file__))
    """
    if not head:
        head_path = os.path.join(path, ".git", "HEAD")
        if not os.path.exists(head_path):
            raise Exception(f"Cannot identify HEAD for git repository at {path}")

        with open(head_path) as fp:
            head = str(fp.read()).strip()

        if head.startswith("ref: "):
            head = head[5:]
            revision_file = os.path.join(path, ".git", *head.split("/"))
        else:
            return head
    else:
        revision_file = os.path.join(path, ".git", "refs", "heads", head)

    if not os.path.exists(revision_file):
        if not os.path.exists(os.path.join(path, ".git")):
            raise Exception(f"{path} does not seem to be the root of a git repository")

        # Check for our .git/packed-refs' file since a `git gc` may have run
        # https://git-scm.com/book/en/v2/Git-Internals-Maintenance-and-Data-Recovery
        packed_file = os.path.join(path, ".git", "packed-refs")
        if os.path.exists(packed_file):
            with open(packed_file) as fh:
                for line in fh:
                    line = line.rstrip()
                    if line and line[:1] not in ("#", "^"):
                        try:
                            revision, ref = line.split(" ", 1)
                        except ValueError:
                            continue
                        if ref == head:
                            return str(revision)

        raise Exception(f'Unable to find ref to head "{head}" in repository')

    with open(revision_file) as fh:
        return str(fh.read()).strip()



import os
import urllib.parse
from logging import getLogger

import logfire


class ASGIMiddlewareAccessLogging:
    logfire: bool = False

    def __init__(self, app, remote_addr_header_name: str = "X-Forwarded-For") -> None:
        self.logger = getLogger("ASGI")
        self.app = app
        # X-Forwarded-For/X-Real-IP
        self.remote_addr_header_name = remote_addr_header_name.lower().encode("latin1")

        # init logfire
        logfire_token = os.getenv("LOGFIRE_TOKEN")
        if logfire_token:
            logfire.configure(token=logfire_token)
            self.logfire = True

    async def __call__(self, scope, receive, send) -> None:
        data = dict()
        data["response"] = {"status": 500}

        async def inner_send(message) -> None:
            if message["type"] == "http.response.start":
                data["response"] = message
            await send(message)

        try:
            # info["start_time"] = time.time()
            await self.app(scope, receive, inner_send)
        except Exception as e:
            # info["response"]["status"] = 500

            raise e
        finally:
            # info["end_time"] = time.time()
            pass

        if scope["type"] != "http":
            return

        headers = scope.get("headers")
        headers = dict(headers)

        remote_addr = self._get_remote_addr(scope, headers)
        user_agent = headers.get(b"user-agent", b"").decode("utf-8")

        request_method = scope.get("method")
        request_path = scope.get("path")
        request_query_string = scope.get("query_string")
        if request_query_string:
            request_path += urllib.parse.unquote(request_query_string)

        status_code = data["response"].get("status")

        self.logger.info(
            f'{remote_addr} - "{request_method} {request_path}" {status_code}'
        )

        if self.logfire:
            logfire.info(
                f"{remote_addr} {request_method} {request_path}",
                remote_addr=remote_addr,
                request_method=request_method,
                request_path=request_path,
                status_code=status_code,
                user_agent=user_agent,
            )

    def _get_remote_addr(self, scope, headers: dict) -> str:
        remote_addr = headers.get(self.remote_addr_header_name, b"").decode("latin1")

        if remote_addr:
            return remote_addr

        remote_addr = scope.get("client")
        if remote_addr:
            return f"{remote_addr[0]}:{remote_addr[1]}"

        return "-"



#!/usr/bin/env python


from importlib import import_module


def autodiscover_schedule(installed_app_name_list):
    """auto discover schedule
    from:
    {
        'every_minute': [
            {'schedule': crontab(minute='*/1'), 'task': 'esi_db.tasks.demo.demo'},
        ],
    }
    to:<The key words are at both ends>
    {
        'app_name.very_minute.esi_db.tasks.demo.demo': {'schedule': crontab(minute='*/1'), 'task': 'esi_db.tasks.demo.demo'},
    }
    """
    celery_beat_schedule = dict()

    for app_name in installed_app_name_list:
        try:
            celery = import_module(f"{app_name}.celery")
        except ModuleNotFoundError:
            continue

        for group in celery.schedule:
            for entry in celery.schedule[group]:
                name = "{}.{}.{}".format(app_name, group, entry["task"])
                celery_beat_schedule[name] = entry

    return celery_beat_schedule



from django_vises.django_settings.helpers import parser_database_uri


def test_parser_database_uri():
    assert parser_database_uri("sqlite://") == {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db.sqlite3",
    }

    assert parser_database_uri("sqlite://db_v2.sqlite3") == {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db_v2.sqlite3",
    }

    assert parser_database_uri("sqlite:///data/db.sqlite3") == {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "/data/db.sqlite3",
    }

    assert parser_database_uri("postgresql://user:password@localhost:5432/db") == {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "db",
        "USER": "user",
        "PASSWORD": "password",
        "HOST": "localhost",
        "PORT": 5432,
    }



from pathlib import Path
from typing import Any
from urllib.parse import ParseResult, urlparse

SQLITE_FILE_NAME = "db.sqlite3"


def _parser_sqlite_uri(
    data: ParseResult, base_dir: Path | None = None
) -> dict[str, Any]:
    if data.path:
        sqlite_file_name = data.path

    elif data.hostname:
        sqlite_file_name = data.hostname

    else:
        if isinstance(base_dir, Path):
            sqlite_file_name = base_dir.joinpath(SQLITE_FILE_NAME).as_posix()
        else:
            sqlite_file_name = SQLITE_FILE_NAME

    return {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": sqlite_file_name,
    }


def parser_database_uri(uri: str, base_dir: Path | None = None) -> dict[str, Any]:
    """返回 Django settings DATABASES 字典"""
    data = urlparse(uri)
    match data.scheme:
        case "sqlite":
            return _parser_sqlite_uri(data, base_dir)

        case "postgresql":
            engine = "django.db.backends.postgresql"

        case _:
            raise

    return {
        "ENGINE": engine,
        "HOST": data.hostname,
        "PORT": data.port,
        "NAME": data.path.lstrip("/"),
        "USER": data.username,
        "PASSWORD": data.password,
    }



"""
https://gist.github.com/rexzhang/74936b88e43b928149af4237d70c3fde
"""

from typing import Any, Optional

import sentry_sdk
from sentry_sdk import configure_scope
from sentry_sdk.integrations import Integration

EVENT_RATE_LIMIT_TIME_RANGE = 60 * 60  # one hour
EVENT_RATE_LIMIT_TIMES = 10  # 10 times

Event = Optional[dict[str, Any]]
EventCounter = dict[str, list[int]]  # { key: [timestamp, count] }
event_counter: EventCounter = dict()


class StartupReport(Exception):
    pass


def auto_drop_event_for_rate_limit(event: Event, _) -> Event:
    try:
        values = event["exception"]["values"][0]
        if not isinstance(values, dict):
            raise ValueError

        key = "{}:{}:{}".format(
            values.get("module", "m"),
            values.get("type", "t"),
            values.get("value", "v"),
        )
    except ValueError:
        key = "m:t:v"

    # timestamp = int(datetime.utcnow().timestamp())
    timestamp = 0  # TODO unimplemented; check, take from event.get(timestamp)

    if key not in event_counter:
        # first event
        event_counter[key] = list((timestamp, 1))
        return event

    # count
    event_counter[key][1] += 1

    if event_counter[key][1] > EVENT_RATE_LIMIT_TIMES:
        # rate limit out, skip event
        return None

    if event_counter[key][1] == EVENT_RATE_LIMIT_TIMES:
        # last time
        with configure_scope() as scope:
            scope.set_tag(
                "rate_limit",
                f"{EVENT_RATE_LIMIT_TIMES}@{EVENT_RATE_LIMIT_TIME_RANGE}",
            )
        return event

    return event


def get_mac_address() -> str:
    from uuid import getnode

    mac = ":".join(
        [f"{(getnode() >> ele) & 0xFF:02x}" for ele in range(0, 8 * 6, 8)][::-1]
    ).upper()
    return mac


def init_sentry(
    dsn: str,
    integrations: list[Integration],
    app_name: str = "PyApp",
    app_version: str = "0.0.0",
    debug: bool = False,
    auto_rate_limit: bool = True,
    announce_at_startup: bool = True,
    user_id_is_mac_address: bool = False,
):
    if debug:
        environment = "debug"
        traces_sample_rate = 1.0
    else:
        environment = "release"
        traces_sample_rate = 0.0
        auto_rate_limit = True
        announce_at_startup = True

    sentry_sdk.init(
        dsn=dsn,
        environment=environment,
        # release=f"{app_name}@{app_version}",
        integrations=integrations,
        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for performance monitoring.
        # We recommend adjusting this value in production,
        traces_sample_rate=traces_sample_rate,
        # rate limit
        before_send=auto_drop_event_for_rate_limit if auto_rate_limit else None,
    )

    mac_address = get_mac_address()
    if user_id_is_mac_address:
        with configure_scope() as scope:
            scope.set_user({"id": mac_address})

    if announce_at_startup:
        sentry_sdk.capture_exception(
            StartupReport(f"{app_name} v{app_version}@{mac_address} is up.")
        )



from uuid import UUID

import redis


class DataSetRedisAbs:
    def __init__(self, url: str, name: str):
        self.conn = redis.from_url(url=url)
        self.name = name

    def add(self, item: bytes):
        self.conn.sadd(self.name, item)

    def pop(self):
        raise

    def remove(self, item: bytes):
        self.conn.srem(self.name, item)

    def all(self) -> set[bytes]:
        return self.conn.smembers(self.name)

    def purge(self):
        return self.conn.delete(self.name)


class DataSetInt(DataSetRedisAbs):
    """
    int.to_bytes(length, byteorder, *, signed=False)
    需要指定转换的长度，这导致通用型变差，不如统一转换成 str 再转换为 bytes
    """

    def add(self, item: int | str):
        if isinstance(item, int):
            item = str(item)

        super().add(bytes(item, "utf-8"))

    def remove(self, item: int | str):
        if isinstance(item, int):
            item = str(item)

        super().remove(bytes(item, "utf-8"))

    def all(self) -> list[str]:
        return [str(item, "utf-8") for item in super().all()]


class DataSetUUID(DataSetRedisAbs):
    def add(self, item: UUID | str):
        if isinstance(item, str):
            item = UUID(item)

        super().add(item.bytes)

    def remove(self, item: UUID | str):
        if isinstance(item, str):
            item = UUID(item)

        super().remove(item.bytes)

    def all(self) -> list[UUID]:
        return [UUID(bytes=item) for item in super().all()]


if __name__ == "__main__":
    r = DataSetInt(url="redis://localhost:6379/2", name="test")
    r.purge()
    r.add(1)
    print(r.all())
    r.add("2")
    print(r.all())
    r.remove("2")
    print(r.all())
    r.remove(1)
    print(r.all())

    from uuid import uuid4

    r = DataSetUUID(url="redis://localhost:6379/2", name="test")
    r.purge()
    r.add(uuid4())
    print(r.all())
    r.add(uuid4().hex)
    print(r.all())
